package FixLoopsWithModeller;

use strict;
use warnings;

BEGIN {
	use Exporter ();
        our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
	$VERSION = "0.1";
	#$DATE    = "02-19-2015";
        #$AUTHOR  = "Brinda Vallat (brinda\@fiserlab.org )";
        @ISA = qw(Exporter);
	
        # name of the functions to export
	@EXPORT = qw(remodel renumber_pdb get_CA_only_chains get_gap_list);

	@EXPORT_OK   = qw(
        );  # symbols to export on request
}

use Data::Dumper;
use Carp;

our @EXPORT_OK;
our $DEBUG = 0;
use PDBfileParser qw(locate_pdb_file);

# This declares the named variables as package globals in the current package.
our $MODELLER_LOG_FILES; 
our $MODELLER_OUT_FILES; 
our $MODELLER_PIR_FILES;
our $MODELLER_PATH;
our $MODELLER_EXEC;
our $EMBOSS_PATH;
our $NEEDLE_EXEC;

=for

CHANGES to this version: 
New modules added:  check_restype
                    check_modres_for_chromophores
                    match_in_array
Changed modules: get_modified_residues
                 get_seqres
                 translate
=cut

sub load_pdb_file {

	my %args = (
        pdb                => '',
        chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        @_,
    );

    my $pdb                = $args{'pdb'}                || undef;
    my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;

	#print "$pdb\t$chain\t$pdb_path\t$pdb_obsoletes_path\n";

    croak "pdb_code is required"           unless $pdb;
    croak "chain is required"              unless defined $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;

    $pdb   = lc $pdb;

	my $filename = PDBfileParser::locate_pdb_file(
        pdb_obsoletes_path => $pdb_obsoletes_path,
        pdb_path           => $pdb_path,
        pdb_code           => $pdb,
        chain              => $chain,
    );

    # open pdb file
	my $fh;
     if ( $filename =~ /\.gz$/ ) { 
            open $fh, '-|', 'gzip', '-dc', $filename or croak "can't open file $filename $!";
     } else {
        open $fh, "<", $filename or croak "can't open file $filename $!";
     }

    my @pdb_lines = <$fh>;
    chomp @pdb_lines;
    close $fh;

    return @pdb_lines;
}

=for get_CA_only_chains
    
    Identify if a pdb file has C-alpha only chains
   
    input : 4-letter pdb code
    output: array of C-alpha only chains 

pdb1ml5.ent  
MDLTYP    CA ATOMS ONLY, CHAIN Z, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S,
MDLTYP   2 T, U, V, W, X, C, D, E, F, G, H, L, M, N, O, P, Q, R, S, T, U, V, W,
MDLTYP   3 X; P ATOMS ONLY, CHAIN A, A, B
 

=cut
sub get_CA_only_chains {
    my %args = (
        pdb                => '',
		chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        @_,
    );
    
    my $pdb                = $args{'pdb'}                || undef;
	my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;

	#print "$pdb\t$chain\t$pdb_path\t$pdb_obsoletes_path\n";

    croak "pdb is required"                unless $pdb;
	croak "chain is required"              unless defined $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;
    
	#print "$pdb\t$chain\t$pdb_path\t$pdb_obsoletes_path\n";

    $pdb   = lc $pdb;
    my @pdb_lines = load_pdb_file( 
		pdb                => $pdb, 
		chain              => $chain,
        pdb_path           => $pdb_path,
        pdb_obsoletes_path => $pdb_obsoletes_path
     );
     
        my @ca_only_chains;   

	my $ca_only = "CA ATOMS ONLY";
	my $mdltype = '';
	my $nn = 0;
	foreach my $newline (@pdb_lines) {
		chomp $newline;
		last if ($newline =~ /^ATOM/ || $newline =~ /^HETATM/);
		
                if ($newline =~ /^MDLTYP/) {
			$nn++;
			my $ss = substr $newline, 10, 70;
			$mdltype = $mdltype.$ss;
		}
	}
       
	if ($nn > 0) {
=for

$mdltype will contain something like this:
 
CA ATOMS ONLY, CHAIN Z, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S,
T, U, V, W, X, C, D, E, F, G, H, L, M, N, O, P, Q, R, S, T, U, V, W,
X; P ATOMS ONLY, CHAIN A, A, B

=cut               
		my @mdl_list = split(';',$mdltype);
		foreach my $mdl_line (@mdl_list) {
			if ($mdl_line =~ /$ca_only/) {
				my @mdl2 = split(',',$mdl_line);
				shift (@mdl2);
				foreach my $chain (@mdl2) {
					$chain =~ s/,//g;
					$chain =~ s/CHAINS*//g;
					$chain =~ s/chains*//g;
					$chain =~ s/Chains*//g;
					$chain =~ s/\s+//g;
                                        push @ca_only_chains, $chain; 
				}
			}
		}
	}
        return @ca_only_chains; 
}

sub get_modified_residues {
    my ( $chain, @pdb_lines ) = @_;

    croak "pdb_lines is required" unless @pdb_lines;
    croak "chain is required"     unless defined $chain;
    my $az = 0;
    my @modres;
    foreach my $newline (@pdb_lines) {

        # print "*** $newline\n";
        chomp $newline;
        if ( $newline =~ /^MODRES/ ) {
            # print "*** $newline\n";
            my $currchain  = substr( $newline, 16, 1 );
            my $modrestype = substr( $newline, 12, 3 );
            my $restype    = substr( $newline, 24, 3 );
            my $resno      = substr( $newline, 18, 5 );
            if ( $currchain eq $chain ) {
                unless ( $modrestype eq $restype ) {
                    my $check = check_restype ($restype);
                    if ($check == 1) {  #This is a natural amino acid
                        push @{ $modres[$az] }, $modrestype, $resno, $restype;
                        $az++;
                    }
                }
            }
        }
    }

    #if ( $az > 0 ) {
    #    print "Warning: Check MODRES: May have non-standard amino acids\n";
    #}

    # print Dumper(\@modres);

    my @new_modres = check_modres_for_chromophores(@modres);

    return @new_modres;
}

=head2 check_restype

    Checks if a given 3-letter residue is a one of the 20 naturally occuring amino acid. 

=cut

sub check_restype {
    my ($restype) = @_; 
        my @aalist = qw(
            ALA
            CYS
            ASP
            GLU
            PHE
            GLY
            HIS
            ILE
            LYS
            LEU
            MET
            ASN
            PRO
            GLN
            ARG
            SER
            THR
            VAL
            TRP
            TYR
        );
    
        my $check = match_in_array ($restype, @aalist);

        return $check;
}

=head2 check_modres_for_chromophores

    Checks if there are inconsistencies in the modified amino acid residues array 
    such as Chromophores. 

=cut

sub check_modres_for_chromophores {
    
    my (@modres) = @_;
    my $num = scalar (@modres);
    my @residues_to_remove;

    FLOOP1: for (my $i = 0; $i < $num; $i++) {
            for (my $j = $i+1; $j < $num; $j++) {
                if ($modres[$i][0] eq $modres[$j][0] && $modres[$i][1] == $modres[$j][1]) {
                    if ($modres[$i][2] ne $modres[$j][2]) { #This is a chromophore or some complicated modification
                        my $check0 = match_in_array($modres[$i][1],@residues_to_remove);
                            if ($check0 == 0) {  #This is a new chromophore
                                push (@residues_to_remove, $modres[$i][1]);
                            }
                        next FLOOP1;
                    }
                }
            }
        }

        my $az = 0;
        my @new_modres;

        FLOOP2: for (my $k = 0; $k < $num; $k++) {
                    my $check2 = match_in_array($modres[$k][1],@residues_to_remove);
                        if ($check2 == 1) {
                            next FLOOP2;
                        }
                    push @{ $new_modres[$az] }, $modres[$k][0], $modres[$k][1], $modres[$k][2];
                    $az++;
                }

        return (@new_modres);

}

=head2 match_in_array

    Checks if a given variable is part of an array

=cut

sub match_in_array {
  my ($query,@array)=@_;

    foreach (@array) {
        if ($query eq $_) {
            return 1;
        }
    }
    return 0;
}

=head2
   Get SEQRES sequence.
   Translate 3-letter aa code to one-letter aa code. 
   UNK aa will be translated to X.
=cut
sub get_seq_res {
    my ( $chain, $aref_modres, $aref_pdb_lines ) = @_;

    croak "pdb_lines is required" unless @$aref_pdb_lines;
    croak "chain is required"     unless defined $chain;
    my $seq = "";
    my $seqlen;

    my @modres = @$aref_modres if @$aref_modres;
    foreach my $line (@$aref_pdb_lines) {
        chomp $line;
        if ( $line =~ /^SEQRES/ ) {
            my @lin = split( '\s+', $line );
            if ( $lin[2] eq $chain ) {
                $seqlen = $lin[3];
                for ( my $aa = 4 ; $aa < scalar(@lin) ; $aa++ ) {
                    #my $a1 = translate( $lin[$aa] );
                    #if ( $a1 eq 'X' ) {
                        my $a2 = check_modres( $lin[$aa], @modres );
                        my $a1 = translate($a2);
                        #print "a2 = $a2 a1 = $a1\n";
                    #}
                    $seq = $seq . $a1;
                }
            }
        }
    }
    # print "seq = $seq\n";
    return $seq;
}

sub write_local_pdb {
    my ( $localpdb, $chain, $aref_pdb_lines, $aref_modres ) = @_;

    croak "pdb_lines is required" unless @$aref_pdb_lines;

    # croak "modres is required\n"               unless @$aref_modres;
    croak "destination pdb file is required" unless $localpdb;

    open PDB, ">$localpdb" or croak "could not open $localpdb for writing";

    my @modres = @$aref_modres if @$aref_modres;
  LINE: foreach my $pdbline (@$aref_pdb_lines) {
        chomp $pdbline;
        if ( $pdbline =~ /^ENDMDL/ ) {
            last LINE;
        }
        if ( $pdbline =~ /^ATOM/ ) {
            my $currentchain = substr( $pdbline, 21, 1 );
            if ( $currentchain eq $chain ) {
                my $newpdbline3 = modline( $pdbline, @modres );
                print PDB "$newpdbline3\n";
            }

        }
        elsif ( $pdbline =~ /^HETATM/ ) {
            my $currentchain = substr( $pdbline, 21, 1 );
            if ( $currentchain eq $chain ) {
                my $newpdbline4 = modline( $pdbline, @modres );
                print PDB "$newpdbline4\n";
            }
        }
    }
    close PDB or croak "could not close";
    return 1;
}

=for
http://salilab.org/archives/modeller_usage/2012/msg00000.html

	Modeller always generates output files in the current directory. 
	So if you want to generate outputs in /home/xxx/output/, 
	change into that directory first before running your script. 
	e.g. if your script is /home/xxx/input/foo.py, 
	run it with something like

	cd /home/xxx/output
	python ../input/foo.py

	Ben Webb, Modeller Caretaker
=cut
# sub fix_loops {
# structure          => whole || loop "  
sub remodel {
    use File::Copy;
    
    my %args = (
        pdb                => '',
        chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        structure          => '', 
        @_,
    );
    my $pdb                = $args{'pdb'}                || undef;
    my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;
    my $structure          = $args{'structure'}          || 'loop';

    croak "pdb is required"                unless $pdb;
    croak "chain is required"              unless defined $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;
    croak "structure invalid parameter"    unless ($structure eq "whole" or $structure eq "loop"); 
    
    make_path($MODELLER_LOG_FILES) unless ( -d $MODELLER_LOG_FILES );
    make_path($MODELLER_OUT_FILES) unless ( -d $MODELLER_OUT_FILES );
    make_path($MODELLER_PIR_FILES) unless ( -d $MODELLER_PIR_FILES );

    # open pdb file
    my @pdb_lines = load_pdb_file(
        pdb                => $pdb,
        chain              => $chain,
        pdb_path           => $pdb_path,
        pdb_obsoletes_path => $pdb_obsoletes_path
    );

    #### Find out if any modified amino acid residues exist
    my @modres = get_modified_residues( $chain, @pdb_lines );

    ### Get SEQRES sequence
    my $sseq = "";                                            # not used
    my $seq = get_seq_res( $chain, \@modres, \@pdb_lines );
    # print "seq2 = $seq\n";

    ### Write pdb file locally - helps with modeller
    my $localpdb = "$pdb$chain.pdb";
    print "Working on\t$pdb\t$chain\t$localpdb\n";
    write_local_pdb( $localpdb, $chain, \@pdb_lines, \@modres );

    ####Write modeller python script for getting ATOM record sequence from pdb file

    my $file1   = "$pdb.$chain.1.seq";
    my $modfile = "$pdb.$chain.sequence.py";
    open MOD, ">$modfile" or croak "Cannot open modfile";
    print MOD "from modeller import *\n";
    print MOD "log.verbose()\n";
    print MOD "env = environ()\n";
    print MOD "env.io.atom_files_directory = ['.']\n";
    print MOD "mdl = model(env)\n";
    print MOD "mdl.read(file='$pdb$chain')\n";
    print MOD "aln = alignment(env)\n";
    print MOD "aln.append_model(mdl, align_codes='$pdb$chain')\n";
    print MOD "aln.write(file='$file1',alignment_format='FASTA')\n";
    close(MOD);

    ###Run Modeller to get sequence
    # `/apps1/modeller/9.9/bin/mod9.9 $modfile > ./modeller_log_files/$pdb.$chain.sequence.log`;
    # my $cmd = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $modfile > ". $MODELLER_LOG_FILES. "/$pdb.$chain.sequence.log";
    my $cmd = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $modfile > q 2>&1";
    system($cmd);
    
    # We move log files to log directoy because Modeller 
    # always generates output files in the current directory  
    if ( -e "$pdb.$chain.sequence.log") {
        move( "$pdb.$chain.sequence.log", $MODELLER_LOG_FILES) or die "Move failed: $!";
    }
	#if ($seq eq $sseq) {
	#	print  "Warning: SEQRES and ATOM record sequences match Modeller may not be needed\n";
	#unlink "$localpdb";
	#}

    # my $file1 = "$pdb.$chain.1.seq";
    unless ( -e $file1 ) {
        print "$pdb\t$chain\tError: Modeller unable to get sequence\n";
        clean_tmp_files( $pdb, $chain ) if not $DEBUG;
        croak "$pdb\t$chain\tError: Modeller unable to get sequence";
    }

    ###Proceed with Modeller
    ###Set up variables for modeller
    my $tt      = $pdb . $chain;
    my $tt_fill = $pdb . $chain . "_fill";
    my $file2   = "$pdb.$chain.2.seq";
    my $file0   = "$pdb.$chain.12.needle";
    my $file3   = "$pdb.$chain.12.ali";
    my $file4   = "$pdb.$chain.pir.ali";
    my $file5   = "$pdb.$chain.loopmodel.py";
    my $file6   = "$pdb.$chain.superpose.py";
    my $file7   = "$tt_fill.B99990001.pdb";
    my $file8   = "$tt_fill.pdb";
    my $n2      = length($seq);
    my $n1      = length($sseq);

    #print "$n1\t$n2\n";

    ###Get SEQRES sequence file in FASTA format

    open OUT2, ">$file2" or croak "Cannot open outfile2";
    print OUT2 ">$tt_fill\n";
    print OUT2 "$seq\n";
    close OUT2;

    ###Run Needleman and Wunsch global alignment algorithm from EMBOSS (needle)
# `~/brinda/software/EMBOSS-6.6.0/emboss/needle -asequence $file1 -bsequence $file2 -gapopen 10 -gapextend 0.5 -outfile $file0 -aformat3 markx3 -awidth3 60`;

    my $needle_cmd =
        $EMBOSS_PATH . "/"
      . $NEEDLE_EXEC
      . " -asequence $file1 -bsequence $file2 -gapopen 10 -gapextend 0.5 -outfile $file0 -aformat3 markx3 -awidth3 60 -sprotein1 -sprotein2 >q 2>&1";
    system($needle_cmd);
    die "$pdb\t$chain\tError: Apparently NEEDLE crashed" unless ( -e $file0 );

    ###Clean the EMBOSS ouptut

    `sed -n '/#/ !p' $file0 | sed '/^\$/d' > $file3`;

    ####Get Alignment in Modeller format

	##Get atom and seqres sequence from needle alignment output
	my $atom_seq = concatenate($file3,1);
	my $seqres_seq = concatenate($file3,2);

	##Find gaps in atom sequence
	my (@gap_list) = find_gaps($atom_seq);	
        #croak "" unless @gap_list; 
	
	##Delete C-terminal overhangs in both atom and seqres sequences
	my $new_atom_seq   = delete_C_term_overhang($atom_seq,@gap_list);
	my $new_seqres_seq = delete_C_term_overhang($seqres_seq,@gap_list);

	##Delete N-terminal overhangs in both atom and seqres sequences
	my $new2_atom_seq   = delete_N_term_overhang($new_atom_seq,@gap_list);
	my $new2_seqres_seq = delete_N_term_overhang($new_seqres_seq,@gap_list);

	##Write alignment in PIR format for Modeller
	my $id=$tt;
	write_pir_file($file4,$id,$chain,$new2_atom_seq,$new2_seqres_seq);

	##Get new gap list after N and C terminal overhangs are deleted
	my (@new_gap_list) = find_gaps($new2_atom_seq);

    ###Get first residue number (not used anymore due to change in policy)

    	#my $first_rn = getFirstResidueNumber($localpdb);

    ####Get Modeller python script for modeling
    if ($structure eq 'loop') {
	    open OUT5, ">$file5" or croak "Cannot open outfile5";
	    print OUT5 "from modeller import *\n";
	    print OUT5 "from modeller.automodel import *\n";
	    print OUT5 "log.verbose()\n";
	    print OUT5 "env = environ()\n";
	    print OUT5 "env.io.atom_files_directory = ['.']\n";
	    print OUT5 "class MyModel(automodel):\n";
	    print OUT5 "\tdef special_patches(self, aln):\n";
	    print OUT5 "\t\tself.rename_segments(segment_ids=['$chain'],renumber_residues=[1])\n";
	    print OUT5 "\tdef select_atoms(self):\n";
	    print OUT5 "\t\treturn selection(\n";
	    for (my $bb=0;$bb<scalar(@new_gap_list);$bb++) {
	    print OUT5 "\t\t\tself.residue_range('$new_gap_list[$bb][0]:$chain','$new_gap_list[$bb][1]:$chain'),\n";
	    }
	    print OUT5 "\t\t)\n";
	    print OUT5 "a = MyModel(env,\n";
	    print OUT5 "\talnfile  = '$file4',\n";
	    print OUT5 "\tknowns   = '$tt',\n";
	    print OUT5 "\tsequence = '$tt_fill')\n";
	    print OUT5 "a.starting_model= 1\n";
	    print OUT5 "a.ending_model  = 1\n";
	    print OUT5 "a.make()\n";
	    close(OUT5);
    } elsif ( $structure eq 'whole') {
	    open OUT5, ">$file5" or croak "Cannot open outfile5";
	    print OUT5 "from modeller import *\n";
	    print OUT5 "from modeller.automodel import *\n";
	    print OUT5 "log.verbose()\n";
	    print OUT5 "env = environ()\n";
	    print OUT5 "env.io.atom_files_directory = ['.']\n";
	    print OUT5 "class MyModel(automodel):\n";
	    print OUT5 "\tdef special_patches(self, aln):\n";
	    print OUT5 "\t\tself.rename_segments(segment_ids=['$chain'],renumber_residues=[1])\n";
	    print OUT5 "a = MyModel(env,\n";
	    print OUT5 "\talnfile  = '$file4',\n";
	    print OUT5 "\tknowns   = '$tt',\n";
	    print OUT5 "\tsequence = '$tt_fill')\n";
	    print OUT5 "a.starting_model= 1\n";
	    print OUT5 "a.ending_model  = 1\n";
	    print OUT5 "a.make()\n";
	    close(OUT5);
    } else {
	    croak "Invalid structure parameter for modeling $structure";
    }

    ###Run Modeller

    #my $cmd_mod2 = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $file5 > ". $MODELLER_LOG_FILES. "/$pdb.$chain.model.log";
    my $cmd_mod2 = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $file5  > q 2>&1";
    system($cmd_mod2);
    
    if ( -e "$pdb.$chain.model.log" ){
        move( "$pdb.$chain.model.log", $MODELLER_LOG_FILES) or die "Move failed: $!";
    }
    unless ( -e $file7 ) {
        print "$pdb\t$chain\tError: Modeller modeling failed: No output\n";
        clean_tmp_files( $pdb, $chain ) if not $DEBUG;
        croak "$pdb\t$chain\tError: Modeller modeling failed: No output";
    }

    copy( $file7, $file8 ) or die "Copy failed: $!";

    # my $cmd_mv = "mv $file7 ". $MODELLER_OUT_FILES. "/pdb$pdb$chain.ent";

    my $MODEL_FILE  = "pdb$pdb$chain.ent";
    my $destination = $MODELLER_OUT_FILES . "/" . $MODEL_FILE;
    move( $file7, $destination ) or die "Move failed: $!";
    print "$pdb\t$chain\tModeller modeling complete\n";

    ###Get Modeller python script for superposition

    open OUT6, ">$file6" or croak "Cannot open outfile6";
    print OUT6 "from modeller import *\n";
    print OUT6 "log.verbose()\n";
    print OUT6 "env = environ()\n";
    print OUT6 "env.io.atom_files_directory = ['.']\n";
    print OUT6 "mdl = model(env, file='$tt')\n";
    print OUT6 "mdl2 = model(env, file='$tt_fill')\n";
    print OUT6
      "aln = alignment(env, file='$file4', align_codes=('$tt', '$tt_fill'))\n";
    print OUT6 "atmsel = selection(mdl)\n";
    print OUT6 "r = atmsel.superpose(mdl2, aln)\n";
    print OUT6 "rms = r.rms\n";
    print OUT6 'print("%d RMSD" % r.rms)', "\n";
    close(OUT6);

    ##Run Modeller superposition

    # my $cmd_mod3 = $MODELLER_PATH . "/". $MODELLER_EXEC. " $file6 > ". $MODELLER_LOG_FILES. "/$pdb.$chain.superpose.log";
    my $cmd_mod3 = $MODELLER_PATH . "/". $MODELLER_EXEC. " $file6  >q 2>&1";
    system($cmd_mod3);
    
    if ( -e "$pdb.$chain.superpose.log") {
        move( "$pdb.$chain.superpose.log", $MODELLER_LOG_FILES) or die "Move failed: $!";
    }
    
    clean_tmp_files( $pdb, $chain ) if not $DEBUG;
    croak "$pdb\t$chain\tError: Modeller superposition failed"
        unless ( -e $MODELLER_LOG_FILES . "/$pdb.$chain.superpose.log" );

    print "$pdb $chain DONE\n";
    print "$destination\n";

    return $destination;

}

=head2 unknown_residues 
	Identifies sequences (SEQRES) that have only unknown residues. 
=cut
sub only_unknown_residues {
    use File::Copy;
    
    my %args = (
        pdb                => '',
        chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        @_,
    );
    my $pdb                = $args{'pdb'}                || undef;
    my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;

    croak "pdb is required"                unless $pdb;
    croak "chain is required"              unless defined  $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;

    # open pdb file
    my @pdb_lines = load_pdb_file(
        pdb                => $pdb,
        chain              => $chain,
        pdb_path           => $pdb_path,
        pdb_obsoletes_path => $pdb_obsoletes_path
    );

    #### Find out if any modified amino acid residues exist
    my @modres = get_modified_residues( $chain, @pdb_lines );

    ### Get SEQRES sequence
    my $seq = get_seq_res( $chain, \@modres, \@pdb_lines );
    
    my $unknown_aa = 0;
    while ( $seq =~ m/X/g ) {
       $unknown_aa++;
    }
    my $seq_length = length $seq;
    
    return 1 if $unknown_aa == $seq_length;
    return 0;

    #my $percentage_unknown = ($unknown_aa/$seq_length)*100;   
 
    #if ( $percentage_unknown >= 50 ) {
    #    return 1; 
    #} else {
    #    return 0; 
    #}
}

=head2 check_mismatch
    Routine to check alignment consistency
    AA residue should be aligned to itself, gap or X.
    if mismatch are found return 1 otherwise 0
=cut
sub check_mismatch {
        my ($seq1, $seq2) = @_;

        die "Sequence 1 is required" unless $seq1;
        die "Sequence 2 is required" unless $seq2;

        $seq1 = uc $seq1;
        $seq2 = uc $seq2;

        my @aa1 = split( //, $seq1 );
        my @aa2 = split( //, $seq2 );

        die "Sequences are of different lengths"
            unless ( scalar(@aa1) == scalar(@aa2) );

        my $mismatch = 0;
        for ( my $i= 0; $i < scalar(@aa1); $i++ ) {

                if ( $aa1[$i] eq $aa2[$i]) {
                        $mismatch = 0;
                }
                elsif ($aa1[$i] eq '-' || $aa2[$i] eq '-') {
                        $mismatch = 0;
                }
                elsif ($aa1[$i] eq 'X' || $aa2[$i] eq 'X') {
                        $mismatch = 0;
                }
                else {
                        $mismatch = 1;
                        last;
                }
        }
        return $mismatch;
}


=head2  get_gap_list
    find missing residues from needleman alignment
    It will return an array with a list of residues containing gaps
    otherwise an empty array is returned.
    
=cut
sub get_gap_list {
    use File::Copy;
    
    my %args = (
        pdb                => '',
        chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        @_,
    );
    my $pdb                = $args{'pdb'}                || undef;
    my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;

    croak "pdb is required"                unless $pdb;
    croak "chain is required"              unless defined  $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;

    # open pdb file
    my @pdb_lines = load_pdb_file(
        pdb                => $pdb,
        chain              => $chain,
        pdb_path           => $pdb_path,
        pdb_obsoletes_path => $pdb_obsoletes_path
    );

    #### Find out if any modified amino acid residues exist
    my @modres = get_modified_residues( $chain, @pdb_lines );

    ### Get SEQRES sequence
    my $sseq = "";                                            # not used
    my $seq = get_seq_res( $chain, \@modres, \@pdb_lines );

    ### Write pdb file locally - helps with modeller
    my $localpdb = "$pdb$chain.pdb";
    print "Working on\t$pdb\t$chain\t$localpdb\n";
    write_local_pdb( $localpdb, $chain, \@pdb_lines, \@modres );

    ### Write modeller python script for getting ATOM record sequence from pdb file

    my $file1   = "$pdb.$chain.1.seq";
    my $modfile = "$pdb.$chain.sequence.py";
    open MOD, ">$modfile" or croak "Cannot open modfile";
    print MOD "from modeller import *\n";
    print MOD "log.verbose()\n";
    print MOD "env = environ()\n";
    print MOD "env.io.atom_files_directory = ['.']\n";
    print MOD "mdl = model(env)\n";
    print MOD "mdl.read(file='$pdb$chain')\n";
    print MOD "aln = alignment(env)\n";
    print MOD "aln.append_model(mdl, align_codes='$pdb$chain')\n";
    print MOD "aln.write(file='$file1',alignment_format='FASTA')\n";
    close(MOD);

    ### Run Modeller to get sequence
    # my $cmd = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $modfile > ". $MODELLER_LOG_FILES. "/$pdb.$chain.sequence.log";
    my $cmd = $MODELLER_PATH . "/" . $MODELLER_EXEC. " $modfile > q 2>&1";
    system($cmd);
    
    # my $file1 = "$pdb.$chain.1.seq";
    unless ( -e $file1 ) {
        print "$pdb\t$chain\tError: Modeller unable to get sequence\n";
        clean_tmp_files( $pdb, $chain ) if not $DEBUG;
        croak "$pdb\t$chain\tError: Modeller unable to get sequence";
    }

    ### Proceed with Modeller
    ### Set up variables for modeller
    my $tt      = $pdb . $chain;
    my $tt_fill = $pdb . $chain . "_fill";
    my $file2   = "$pdb.$chain.2.seq";
    my $file0   = "$pdb.$chain.12.needle";
    my $file3   = "$pdb.$chain.12.ali";
    my $n2      = length($seq);
    my $n1      = length($sseq);

    #print "$n1\t$n2\n";

    ### Get SEQRES sequence file in FASTA format

    open OUT2, ">$file2" or croak "Cannot open outfile2";
    print OUT2 ">$tt_fill\n";
    print OUT2 "$seq\n";
    close OUT2;

    # Run Needleman and Wunsch global alignment algorithm from EMBOSS (needle)
    my $needle_cmd =
        $EMBOSS_PATH . "/"
      . $NEEDLE_EXEC
      . " -asequence $file1 -bsequence $file2 -gapopen 10 -gapextend 0.5 -outfile $file0 -aformat3 markx3 -awidth3 60 -sprotein1 -sprotein2 >q 2>&1";
    system($needle_cmd);
    die "$pdb\t$chain\tError: Apparently NEEDLE crashed" unless ( -e $file0 );

    ### Clean the EMBOSS ouptut

    `sed -n '/#/ !p' $file0 | sed '/^\$/d' > $file3`;

    ### Get Alignment in Modeller format

	# Get atom and seqres sequence from needle alignment output
	my $atom_seq   = concatenate($file3,1);
	my $seqres_seq = concatenate($file3,2);

	## Find gaps in atom sequence
	my (@gap_list) = find_gaps($atom_seq);	
	
	## Delete C-terminal overhangs in both atom and seqres sequences
	my $new_atom_seq   = delete_C_term_overhang($atom_seq,@gap_list);
	my $new_seqres_seq = delete_C_term_overhang($seqres_seq,@gap_list);

	## Delete N-terminal overhangs in both atom and seqres sequences
	my $new2_atom_seq   = delete_N_term_overhang($new_atom_seq,@gap_list);
	my $new2_seqres_seq = delete_N_term_overhang($new_seqres_seq,@gap_list);

	## Get new gap list after N and C terminal overhangs are deleted
	my (@new_gap_list) = find_gaps($new2_atom_seq);

        die "There was error: $pdb $chain. ATOM and SEQRES are different in length" 
            if ( length $new2_atom_seq != length $new2_seqres_seq);
        
        # checking alignment consistency 
        my $mismatch = check_mismatch( $new2_atom_seq, $new2_seqres_seq );

        # Write alignment in PIR format for Modeller
        if ($DEBUG) {
	    my $id    = $tt;
            my $file4 = "temp_brinda.pir";
	    write_pir_file($file4,$id,$chain,$new2_atom_seq,$new2_seqres_seq);
        }
        clean_tmp_files($pdb, $chain) if not $DEBUG;
      
        # If seqres has gaps
        my $gaps_in_seqres = 0;
	if ( $seqres_seq =~ /-/ ) {
            $gaps_in_seqres = 1;
	}
        return ( $mismatch, $gaps_in_seqres, \@new_gap_list) ;
}

# ****************
# SUBROUTINE to clean temporary files
sub clean_tmp_files {
    my ( $pdb, $chain ) = @_;

    my $tt       = $pdb . $chain;
    my $tt_fill  = $pdb . $chain . "_fill";
    my $file1    = "$pdb.$chain.1.seq";
    my $file2    = "$pdb.$chain.2.seq";
    my $file0    = "$pdb.$chain.12.needle";
    my $file3    = "$pdb.$chain.12.ali";
    my $file4    = "$pdb.$chain.pir.ali";
    my $file5    = "$pdb.$chain.loopmodel.py";
    my $file6    = "$pdb.$chain.superpose.py";
    my $file7    = "$tt_fill.B99990001.pdb";
    my $file8    = "$tt_fill.pdb";
    my $localpdb = "$tt.pdb";
    my $modfile  = "$pdb.$chain.sequence.py";

    unlink $file0 if -e $file0;
    unlink $file1 if -e $file1;
    unlink $file2 if -e $file2;
    unlink $file3 if -e $file3;

    #unlink $file4;
    unlink $file5 if -e $file5;
    unlink $file6 if -e $file6;
    unlink $file8 if -e $file8;

    if ( -e $file4 ){ 
        move( $file4, $MODELLER_PIR_FILES . "/" ) or die "Move failed: $!";
    }
    unlink "$tt_fill.ini" if -e "$tt_fill.ini";
    unlink "$tt_fill.rsr" if -e "$tt_fill.rsr";
    unlink "$tt_fill.sch" if -e "$tt_fill.sch";
    unlink $localpdb      if -e $localpdb;
    unlink $modfile       if -e $modfile; 
    unlink "$tt_fill.V99990001" if -e "$tt_fill.V99990001";
    unlink "$tt_fill.D00000001" if -e "$tt_fill.D00000001";
}

#SUBROUTINE to convert from 3-letter codes to 1-letter codes##############
sub translate {
    my ($inp) = @_;
    my $out;
    if    ( $inp eq 'ALA' ) { $out = 'A' }
    elsif ( $inp eq 'CYS' ) { $out = 'C' }
    elsif ( $inp eq 'ASP' ) { $out = 'D' }
    elsif ( $inp eq 'GLU' ) { $out = 'E' }
    elsif ( $inp eq 'PHE' ) { $out = 'F' }
    elsif ( $inp eq 'GLY' ) { $out = 'G' }
    elsif ( $inp eq 'HIS' ) { $out = 'H' }
    elsif ( $inp eq 'ILE' ) { $out = 'I' }
    elsif ( $inp eq 'LYS' ) { $out = 'K' }
    elsif ( $inp eq 'LEU' ) { $out = 'L' }
    elsif ( $inp eq 'MET' ) { $out = 'M' }
    elsif ( $inp eq 'ASN' ) { $out = 'N' }
    elsif ( $inp eq 'PRO' ) { $out = 'P' }
    elsif ( $inp eq 'GLN' ) { $out = 'Q' }
    elsif ( $inp eq 'ARG' ) { $out = 'R' }
    elsif ( $inp eq 'SER' ) { $out = 'S' }
    elsif ( $inp eq 'THR' ) { $out = 'T' }
    elsif ( $inp eq 'VAL' ) { $out = 'V' }
    elsif ( $inp eq 'TRP' ) { $out = 'W' }
    elsif ( $inp eq 'TYR' ) { $out = 'Y' }
    elsif ( $inp eq 'MSE' ) { $out = 'M' }
    elsif ( $inp eq 'MEX' ) { $out = 'C' }
    elsif ( $inp eq 'ABU' ) { $out = 'C' }
    elsif ( $inp eq 'UNK' ) { $out = 'X' }
    else                    { $out = '' }
    return $out;
}

##Subroutine to check modified residues in SEQRES
sub check_modres {
    #use Data::Dumper;
    my ( $res, @modres ) = @_;
    #print Dumper(\$res);
    #print Dumper(\@modres);

    my $out = $res;
    for ( my $aa = 0 ; $aa < scalar(@modres) ; $aa++ ) {
        if ( $res eq $modres[$aa][0] ) {
            $out = $modres[$aa][2];
        }
    }
    # print "out = $out\n";
    return $out;
}

##Subroutine to modify pdb lines that contain modified amino acid residue
sub modline {
    my ( $pdbline, @modres ) = @_;
    my $newpdbline  = $pdbline;
    my $newpdbline2 = $pdbline;

    # print "$pdbline\n";
    my $restype = substr( $pdbline, 17, 3 );
    $restype =~ s/\s+//g if $restype;

    my $resno = substr( $pdbline, 22, 5 );
    $resno =~ s/\s+//g if $resno;

    #print "$restype\t$resno\n";
  FLOOP: for ( my $aa = 0 ; $aa < scalar(@modres) ; $aa++ ) {

        #print "$restype\t$modres[$aa][0]\n";
        if ( $restype eq $modres[$aa][0] ) {

            #print "$restype\t$modres[$aa][0]\n";
            ( $newpdbline  = $pdbline ) =~ s/HETATM/ATOM  /g;
            ( $newpdbline2 = $newpdbline ) =~ s/$restype/$modres[$aa][2]/g;
            last FLOOP;
        }
    }
    return $newpdbline2;
}

##Subroutine to get starting residue number of a pdb file
sub getFirstResidueNumber {
    my ($localpdb) = @_;
    open PDB, "<$localpdb";
    my $line = <PDB>;
    chomp $line;
    my $resno;
    if ( $line =~ /^ATOM/ ) {
        $resno = substr( $line, 22, 5 );
        $resno =~ s/\s+//g;
    }
    elsif ( $line =~ /^HETATM/ ) {
        $resno = substr( $line, 22, 5 );
        $resno =~ s/\s+//g;
    }
    else {
        $resno = 1;
    }
    if ( $resno =~ /[a-zA-Z]/ ) { $resno = 1 }
    close PDB;
    return $resno;
}

sub renumber_pdb {
    use File::Path qw(make_path);
    use File::Copy;
 
    my %args = (
        pdb                => '',
        chain              => '',
        pdb_path           => '',
        pdb_obsoletes_path => '',
        @_,
    );
    my $pdb                = $args{'pdb'}                || undef;
    my $chain              = $args{'chain'}              || undef;
    my $pdb_path           = $args{'pdb_path'}           || undef;
    my $pdb_obsoletes_path = $args{'pdb_obsoletes_path'} || undef;

    croak "pdb is required"              unless $pdb;
    croak "chain is required"            unless defined $chain;
    croak "pdb_path is required"           unless $pdb_path;
    croak "pdb_obsoletes_path is required" unless $pdb_obsoletes_path;
    
    make_path($MODELLER_LOG_FILES ) unless ( -d $MODELLER_LOG_FILES );
    make_path($MODELLER_OUT_FILES ) unless ( -d $MODELLER_OUT_FILES );

    #####Open pdb file
    my @pdb_lines = load_pdb_file(
        pdb                => $pdb,
        chain              => $chain,
        pdb_path           => $pdb_path,
        pdb_obsoletes_path => $pdb_obsoletes_path
    );

    #####Get modified amino acid residues
    my @modres = get_modified_residues( $chain, @pdb_lines );

    ####Get Seqres residues
    my $sseq = "";                                            # not used
    my $seq = get_seq_res( $chain, \@modres, \@pdb_lines );

    ###Write pdb file locally -- helps with modeller
    my $localpdb = "$pdb$chain.pdb";
    print "Working on\t$pdb\t$chain\t$localpdb\n";
    write_local_pdb( $localpdb, $chain, \@pdb_lines, \@modres );

    ###Get first residue number
    #my $first_rn = getFirstResidueNumber($localpdb);

    ####Write modeller python script for renumbering PDB

    my $fileA   = "pdb$pdb$chain.ent";
    my $modfile = "$pdb.$chain.renumber.py";
    open MOD, ">$modfile" or die "Cannot open modfile\n";
    print MOD "from modeller import *\n";
    print MOD "log.verbose()\n";
    print MOD "env = environ()\n";
    print MOD "env.io.atom_files_directory = ['.']\n";
    print MOD "mdl = model(env)\n";
    print MOD "mdl.read(file='$pdb$chain')\n";
    print MOD
"mdl.rename_segments(segment_ids=['$chain'],renumber_residues=[1])\n";
    print MOD "mdl.write(file='$fileA', model_format='PDB')\n";
    close(MOD);

    ###Run Modeller to get renumbered PDB

    # my $cmd = $MODELLER_PATH . "/". $MODELLER_EXEC. " $modfile > ". $MODELLER_LOG_FILES. "/$pdb.$chain.renumber.log";
    # my $cmd = $MODELLER_PATH . "/". $MODELLER_EXEC. " $modfile > /dev/null 2>&1";
    my $cmd = $MODELLER_PATH . "/". $MODELLER_EXEC. " $modfile > q 2>&1";
    system($cmd);

 
    if ( -e "$pdb.$chain.renumber.log") {
        move( "$pdb.$chain.renumber.log", $MODELLER_LOG_FILES) or die "Move failed: $!";
    }

    unless ( -e $fileA ) {
        print "$pdb\t$chain\tError: Modeller renumbering failed\n";
        clean_renumber_files( $pdb, $chain ) if not $DEBUG;
        croak "$pdb\t$chain\tError: Modeller renumbering failed";
    }

    my $MODEL_FILE  = "pdb$pdb$chain.ent";
    my $destination = $MODELLER_OUT_FILES . "/" . $MODEL_FILE;
    move( $fileA, $destination ) or die "Move failed: $!";

    print "$pdb\t$chain\tModeller renumber complete\n";
    clean_renumber_files( $pdb, $chain ) if not $DEBUG;

    print "DONE\n";
    return $destination;

}

#SUBROUTINE to clean temporary files
sub clean_renumber_files {
    my ( $pdb, $chain ) = @_;

    my $tt       = $pdb . $chain;
    my $localpdb = "$tt.pdb";
    my $modfile  = "$pdb.$chain.renumber.py";

    unlink $localpdb;
    unlink $modfile;
}

sub concatenate {

	my ($align_file,$seq_num) = @_;

        croak "Sequence number is required" unless $seq_num;
        croak "Alignment file is required" unless $align_file;

        open (IN, "<$align_file") or croak "Unable to open alignment file $align_file\n";
        my $protein_number = 0;
        my $seq='';

        ALINE:while (my $line = <IN>) {
                chomp $line;
                if ($line =~ /^>/) {
                        $protein_number++;
                        next ALINE;
                } else {
                        if ($protein_number == $seq_num) {
                                $seq=$seq.$line;
                        }
                }
        }
        close (IN);
        return ($seq);

}

sub find_gaps {

        my ($seq) = @_;

	croak "Sequence is required" unless $seq;

	##Add "*" to the end of the sequence - needed for identifying C-terminal gaps
	my $st = '*';
        $seq = $seq.$st;

        my $gap_number = 0;
        my $gap_start = 0;
        my $gap_len = 0;
        my $gap_end = 0;
        my @gap_list;

                my @lin = split ('',$seq);
                for (my $aa=0;$aa<scalar(@lin);$aa++) {
                        if ($lin[$aa] eq '-' && $lin[$aa-1] ne '-') {
                                $gap_start = $aa+1;
                                $gap_len = $gap_start;
                        } elsif ($lin[$aa] eq '-' && $lin[$aa-1] eq '-') {
                                $gap_len++;
                        } elsif ($lin[$aa] ne '-' && $lin[$aa-1] eq '-')  {
                                $gap_end = 1;
                        }
                        if ($gap_end == 1) {
                                # gap number, gap start position, gap end position
                                push @{$gap_list[$gap_number]},$gap_start,$gap_len;
                                $gap_number++;
                                $gap_end = 0;
                                $gap_start = 0;
                                $gap_len = 0;
                        }
                }

	return (@gap_list);
}

sub delete_N_term_overhang {

        my ($seq,@gap_list) = @_;

        croak "Sequence is required" unless $seq;

        return $seq unless @gap_list;
        
        my $seq_length = length($seq);

        my $start      = $gap_list[0][0];
        my $end        = $gap_list[0][1];
        my $gap_length = $end-$start+1;
        my $newseq = $seq;

        if ($start == 1) {      ###N-terminal gaps
                my $new_length = $seq_length-$gap_length;
                my $new_start = $start+$gap_length-1;
                $newseq = substr $seq, $new_start, $new_length;
        }
        return ($newseq);
}

sub delete_C_term_overhang {

        my ($seq, @gap_list) = @_;

        croak "Sequence is required" unless $seq;

        return $seq unless @gap_list;
        
        my $seq_length = length($seq);
        my $start      = $gap_list[-1][0];
        my $end        = $gap_list[-1][1];
        my $gap_length = $end-$start+1;
        my $newseq = $seq;

        if ($end == $seq_length) {      ###C-terminal gaps
                my $new_length = $seq_length-$gap_length;
                $newseq = substr $seq, 0, $new_length;
        }

        return ($newseq);
}

sub write_pir_file {

        my ($file4,$id,$chain,$atom_seq,$seqres_seq) = @_;

        croak "Atom sequence is required" unless $atom_seq;
        croak "Seqres sequence is required" unless $seqres_seq;
        croak "ID is required" unless $id;
        croak "Chain id is required" unless defined $chain;
        croak "Pir output file is required" unless $file4;

        open OUT4, ">$file4" or die "Cannot open output file for writing pir alignment";
        print OUT4 ">P1;$id\n";
        print OUT4 "structure:$id: :$chain: :$chain:temp1:organism1:1.90:0.19\n";
        for (my $aa=0;$aa<length($atom_seq)/60;$aa++) {
                print OUT4 substr($atom_seq,$aa*60,60);
                if ($aa*60+60>=length($atom_seq)) {print OUT4 '*'}
                print OUT4 "\n";
        }
        print OUT4 ">P1;$id" . "_fill\n";
        print OUT4 'sequence::', $chain, '::', $chain, '::::: ',"\n";
        for (my $aa=0;$aa<length($seqres_seq)/60;$aa++) {
                print OUT4 substr($seqres_seq,$aa*60,60);
                if ($aa*60+60>=length($seqres_seq)) {print OUT4 '*'}
                print OUT4 "\n";
        }
        close(OUT4);

        return;
}

1; 
